/*
 * Copyright 2021 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package androidx.camera.integration.extensions.validation

import android.content.Context
import android.content.Intent
import android.os.Bundle
import android.view.LayoutInflater
import android.widget.AdapterView
import android.widget.BaseAdapter
import android.widget.ListView
import android.widget.Toast
import androidx.appcompat.app.AppCompatActivity
import androidx.camera.integration.extensions.ExtensionTestType.TEST_TYPE_CAMERAX_EXTENSION
import androidx.camera.integration.extensions.INVALID_LENS_FACING
import androidx.camera.integration.extensions.IntentExtraKey.INTENT_EXTRA_KEY_CAMERA_ID
import androidx.camera.integration.extensions.IntentExtraKey.INTENT_EXTRA_KEY_EXTENSION_MODE
import androidx.camera.integration.extensions.IntentExtraKey.INTENT_EXTRA_KEY_LENS_FACING
import androidx.camera.integration.extensions.IntentExtraKey.INTENT_EXTRA_KEY_REQUEST_CODE
import androidx.camera.integration.extensions.IntentExtraKey.INTENT_EXTRA_KEY_TEST_TYPE
import androidx.camera.integration.extensions.R
import androidx.camera.integration.extensions.TestResultType.TEST_RESULT_NOT_SUPPORTED
import androidx.camera.integration.extensions.validation.CameraValidationResultActivity.Companion.getLensFacingStringFromInt
import androidx.core.app.ActivityCompat

/**
 * Activity to list all supported extension modes of the specified camera device.
 *
 * Clicking a list item will launch the ImageValidationActivity to trigger the image taking and
 * validation flow.
 */
class ExtensionValidationResultActivity : AppCompatActivity() {
    private lateinit var extensionTestResultMap: LinkedHashMap<Int, Pair<Int, String>>
    private lateinit var testResults: TestResults
    private val result = Intent()
    private var lensFacing = INVALID_LENS_FACING
    private lateinit var testType: String
    private lateinit var cameraId: String
    private lateinit var adapter: BaseAdapter
    private val imageValidationActivityRequestCode =
        ImageValidationActivity::class.java.hashCode() % 1000

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        setContentView(R.layout.full_listview)

        testResults = TestResults.getInstance(this)

        testType = intent.getStringExtra(INTENT_EXTRA_KEY_TEST_TYPE)!!
        cameraId = intent.getStringExtra(INTENT_EXTRA_KEY_CAMERA_ID)!!
        lensFacing = intent.getIntExtra(INTENT_EXTRA_KEY_LENS_FACING, INVALID_LENS_FACING)

        extensionTestResultMap =
            testResults.getCameraExtensionResultMap()[Pair(testType, cameraId)]!!

        result.putExtra(INTENT_EXTRA_KEY_TEST_TYPE, testType)
        result.putExtra(INTENT_EXTRA_KEY_CAMERA_ID, cameraId)
        val requestCode = intent.getIntExtra(INTENT_EXTRA_KEY_REQUEST_CODE, -1)
        setResult(requestCode, result)

        supportActionBar?.title = if (testType == TEST_TYPE_CAMERAX_EXTENSION) {
            resources.getString(R.string.camerax_extensions_validator)
        } else {
            resources.getString(R.string.camera2_extensions_validator)
        }
        supportActionBar!!.subtitle = "Camera $cameraId [${getLensFacingStringFromInt(lensFacing)}]"

        val layoutInflater = getSystemService(Context.LAYOUT_INFLATER_SERVICE) as LayoutInflater
        adapter = ExtensionValidationResultAdapter(testType, layoutInflater, extensionTestResultMap)

        val listView = findViewById<ListView>(R.id.listView)
        listView.adapter = adapter

        listView.onItemClickListener =
            AdapterView.OnItemClickListener { _, _, position, _ ->
                if (extensionTestResultMap.values.elementAt(position).first ==
                    TEST_RESULT_NOT_SUPPORTED) {
                    Toast.makeText(this, "Not supported!", Toast.LENGTH_SHORT).show()
                    return@OnItemClickListener
                }

                startCaptureValidationActivity(
                    testType,
                    cameraId,
                    extensionTestResultMap.keys.elementAt(position)
                )
            }
    }

    @Suppress("DEPRECATION")
    @Deprecated("Deprecated in ComponentActivity")
    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?) {
        super.onActivityResult(requestCode, resultCode, data)

        if (requestCode != imageValidationActivityRequestCode) {
            return
        }

        adapter.notifyDataSetChanged()
    }

    private fun startCaptureValidationActivity(testType: String, cameraId: String, mode: Int) {
        val intent = Intent(this, ImageValidationActivity::class.java)
        intent.putExtra(INTENT_EXTRA_KEY_TEST_TYPE, testType)
        intent.putExtra(INTENT_EXTRA_KEY_CAMERA_ID, cameraId)
        intent.putExtra(INTENT_EXTRA_KEY_LENS_FACING, lensFacing)
        intent.putExtra(INTENT_EXTRA_KEY_EXTENSION_MODE, mode)
        intent.putExtra(INTENT_EXTRA_KEY_REQUEST_CODE, imageValidationActivityRequestCode)

        ActivityCompat.startActivityForResult(
            this,
            intent,
            imageValidationActivityRequestCode,
            null
        )
    }
}
